static char rcsid[] = "@(#)$Id: schedule.c,v 2.18 2017/09/10 13:15:31 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.18 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "schedule_time.h"

DEBUG_VAR(Debug,__FILE__,"net");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

volatile int wait_can_signal = 0;

enum action_status no_action_routine(fd,data,now)
     int fd; 
     union action_routine_data         data;
     const struct schedule_timelimit * now;
{
    return action_disable;
}

static enum action_status ANY_ACTION P_((int fd,
					 union action_routine_data data,
					 const struct schedule_timelimit *now))
    UNUSED_FUNCOK;
static enum action_status ANY_ACTION(fd,data,now)
     int                               fd; 
     union action_routine_data         data;
     const struct schedule_timelimit * now;
{
    return action_disable;
}

void no_free_action_data(data)
     union action_routine_data * data;
{
    data->data = NULL;
}

void no_inc_action_data_refcount(data)
     union action_routine_data data;
{

}

enum schedule_return no_schedule_routine(fd,data,now)
     int                         fd; 
     union schedule_routine_data data;
     const struct schedule_timelimit * now;
{
    return schedule_done;
}

void no_free_schedule_data(data)
     union schedule_routine_data * data;
{
    data->data = NULL;
}

void no_inc_schedule_data_refcount(data)
     union schedule_routine_data data;
{

}

static struct actions {
    
    int    fd; 
    int    timeout_sec;
    action_routine             * read_act; 
    action_routine             * write_act; 		       
    action_routine             * timeout_act;
    free_action_data_f         * free_action_data;
    inc_action_data_refcount_f * inc_action_data_refcount;
    union action_routine_data    data;

    struct schedule_timelimit next_timeout;
    int    busy;                     /* Prevent nested calls --
				      * when read_act(), write_act() or
				      * timeout_act() calls wait_for_timeout()
				      * or wait_for_action() directly or
				      * indirectly (for example with lib_error)
				      */

    schedule_routine             * schedule_act;
    free_schedule_data_f         * free_schedule_data;
    inc_schedule_data_refcount_f * inc_schedule_data_refcount; 
    union schedule_routine_data    schedule_data;

}         * actions       = NULL;
static int  actions_count = 0; 

static void zero_actions_item  P_((struct actions *item));
static void zero_actions_item(item)
     struct actions *item;
{
    item->fd                = -1;
    item->timeout_sec       = 0;
    item->next_timeout      = NO_schedule_timelimit;
    item->busy              = 0;
    item->timeout_sec       = 0;
    item->read_act          = no_action_routine;
    item->write_act         = no_action_routine;
    item->timeout_act       = no_action_routine;
    item->free_action_data         = no_free_action_data;
    item->inc_action_data_refcount = no_inc_action_data_refcount;
    item->data                     = NO_action_routine_data;
    
    item->schedule_act               = no_schedule_routine;
    item->free_schedule_data         = no_free_schedule_data;
    item->inc_schedule_data_refcount = no_inc_schedule_data_refcount;
    item->schedule_data = NO_schedule_routine_data;	
}


static void remove_action P_((int ptr));
static void remove_action(ptr)
     int ptr;
{
    int i;

    if (ptr < 0 || ptr >=  actions_count)
	panic("SCHEDULE PANIC",__FILE__,__LINE__,
	      "remove_action",
	      "Bad index",0);
    
    DPRINT(Debug,4,(&Debug,
		    "** Removing action %d  (fd=%d)\n", ptr,
		    actions[ptr].fd));

    if (actions[ptr].data.data) {
	 DPRINT(Debug,12,(&Debug,"Freeing action %d data (fd=%d)\n",
			  ptr, actions[ptr].fd));
	 actions[ptr].free_action_data(& (actions[ptr].data));
    }

    if (actions[ptr].schedule_data.data) {
	 DPRINT(Debug,12,(&Debug,"Freeing action %d schedule data (fd=%d)\n",
			  ptr, actions[ptr].fd));
	 actions[ptr].free_schedule_data(& (actions[ptr].schedule_data));
    }
    
    zero_actions_item(&(actions[ptr]));
    
    for (i = ptr+1; i < actions_count; i++) {
	actions[i-1] = actions[i];
	zero_actions_item(& (actions[i]));
    }
    actions_count--;
}

const union action_routine_data   NO_action_routine_data;
const union schedule_routine_data NO_schedule_routine_data;

void change_action (fd,timeout_sec,read_act,write_act,timeout_act,
		    free_action_data,inc_action_data_refcount,
		    data)
     int                       fd; 
     int                       timeout_sec;
     action_routine          * read_act; 
     action_routine          * write_act; 	
     action_routine          * timeout_act;
     free_action_data_f      * free_action_data;
     inc_action_data_refcount_f * inc_action_data_refcount;
     union action_routine_data data;
{
    int ptr;

    static int report_time_MAX = 1;

    for (ptr = 0; ptr < actions_count; ptr++)
	if (actions[ptr].fd == fd)
	    break;

    if (ptr == actions_count) {	

	if (no_action_routine == read_act &&
	    no_action_routine == write_act &&
	    no_action_routine == timeout_act) {
	    DPRINT(Debug,9,(&Debug,
			    "change_action: action for (fd=%d) already cleared\n", 
			    fd)); 
	    return;
	}

	actions = safe_array_realloc(actions, (++actions_count),
				     sizeof (struct actions));

	zero_actions_item(& (actions[ptr]));

	DPRINT(Debug,4,(&Debug,
			"** Adding action %d  (fd=%d)\n", ptr,fd)); 
    }

    if (report_time_MAX) {
	char * X = ctime(&time_MAX);
	
	report_time_MAX = 0;
	
	if (X) {
	    DPRINT(Debug,1,(&Debug,
			    "Using time_MAX = %ld  = %s",
			    (long)time_MAX,X));
	} else {
	    DPRINT(Debug,1,(&Debug,
			    "Using time_MAX = %ld\n",
			    (long)time_MAX));
	}
	DPRINT(Debug,1,(&Debug,"\ntime_t size is %ld bytes\n",
			(long)sizeof(time_t)));
    }
    
    
    if (actions[ptr].timeout_sec != timeout_sec) {
	char * X;

	
	if (! schedule_set_next_timeout(& (actions[ptr].next_timeout),
					timeout_sec)) {
	    
	    DPRINT(Debug,9,(&Debug, 
			    "change_action: action %d (fd=%d), time overflow or no current time\n",
			    ptr,fd)); 
			    
	}

	X =  schedule_timeout_string(& actions[ptr].next_timeout);
	
	if (X) {
	    DPRINT(Debug,9,(&Debug, 
			    "change_action: action %d (fd=%d) timeout %d seconds => %s\n",
			    ptr,fd,timeout_sec,X));
	    free(X);
	    X = NULL;
	}	
    }

    if (actions[ptr].data.data != data.data) {
	if (actions[ptr].data.data) {
	    DPRINT(Debug,9,(&Debug, 
			    "change_action: action %d (fd=%d) freeing previous data\n",
			    ptr,fd));
	    actions[ptr].free_action_data(& (actions[ptr].data));
	}
	actions[ptr].data        = data;
	if (actions[ptr].data.data) {
	    inc_action_data_refcount(actions[ptr].data);
	}
    }
    
    actions[ptr].fd          = fd;
    actions[ptr].timeout_sec = timeout_sec;
    actions[ptr].read_act    = read_act;
    actions[ptr].write_act   = write_act;
    actions[ptr].timeout_act = timeout_act;
    actions[ptr].free_action_data         = free_action_data;
    actions[ptr].inc_action_data_refcount = inc_action_data_refcount;
    
    if (no_action_routine == read_act &&
	no_action_routine == write_act &&
	no_action_routine == timeout_act &&
	actions[ptr].schedule_act == no_schedule_routine) {

	if (actions[ptr].busy) {
	    DPRINT(Debug,4,(&Debug,
			    "** Action %d busy (fd=%d): remove skipped\n", 
			    ptr, actions[ptr].fd)); 
	} else
	    remove_action(ptr);
    }
}

void set_schedule_action (fd,routine,
			  free_schedule_data,inc_schedule_data_refcount,
			  data)
     int                         fd;
     schedule_routine          * routine;
     free_schedule_data_f         * free_schedule_data;
     inc_schedule_data_refcount_f * inc_schedule_data_refcount;
     union schedule_routine_data data;
{
    int ptr;

    for (ptr = 0; ptr < actions_count; ptr++)
	if (actions[ptr].fd == fd)
	    break;

    if (ptr == actions_count) {	

	if (no_schedule_routine == routine) {
	    DPRINT(Debug,9,(&Debug,
			    "set_schedule_action: action for (fd=%d) already cleared\n", 
			    fd)); 
	    return;
	}

	actions = safe_array_realloc(actions, (++actions_count),
				     sizeof (struct actions));

	zero_actions_item(& (actions[ptr]));
	
	DPRINT(Debug,4,(&Debug,
			"** Adding action %d  (fd=%d)\n", ptr,fd)); 
    }


    actions[ptr].fd          = fd;
    actions[ptr].schedule_act  = routine;
    
    if (actions[ptr].schedule_data.data != data.data) {
	if (actions[ptr].schedule_data.data) {
	    DPRINT(Debug,9,(&Debug,"set_schedule_action: Freeing action %d previous schedule data (fd=%d)\n",
			     ptr, actions[ptr].fd));
	    actions[ptr].free_schedule_data(& (actions[ptr].schedule_data));
	}
	
	actions[ptr].schedule_data = data;
	if (actions[ptr].schedule_data.data)
	    inc_schedule_data_refcount(actions[ptr].schedule_data);
    }
    actions[ptr].free_schedule_data         = free_schedule_data;
    actions[ptr].inc_schedule_data_refcount = inc_schedule_data_refcount;
    
    if (no_action_routine == actions[ptr].read_act &&
	no_action_routine == actions[ptr].write_act &&
	no_action_routine == actions[ptr].timeout_act &&
	actions[ptr].schedule_act == no_schedule_routine) {
	
	if (actions[ptr].busy) {
	    DPRINT(Debug,4,(&Debug,
			    "** Action %d busy (fd=%d): remove skipped\n", 
			    ptr, actions[ptr].fd)); 
	} else
	    remove_action(ptr);
    }
}			     

void clear_action(fd)
     int fd;
{
    
    set_schedule_action(fd,no_schedule_routine,
			no_free_schedule_data,
			no_inc_schedule_data_refcount,
			NO_schedule_routine_data);
    change_action(fd,0,
		  no_action_routine,no_action_routine,no_action_routine,
		  no_free_action_data,no_inc_action_data_refcount,
		  NO_action_routine_data);
}


/* -------------------------------------------------------------------- */

#define TIMED_TASK_HANDLE_magic 0xF102
struct timed_task_handle {
    unsigned short magic;     /* TIMED_TASK_HANDLE_magic */

    struct timed_task          * task;
    timed_task_free_f          * free_task;
    inc_timed_task_refcount_f  * inc_task_refcount;
    timed_task_action_f        * task_action;

    struct schedule_timelimit    deadline;
    unsigned int                 fired:1;   /* task_action called         */
    unsigned int                 busy:1;    /* task_action() on execution */
    unsigned int                 delete:1;  /* delete after task_action() */
					    /*  ---                       */
					    /* used if alter_timed_task() */
					    /* called when busy           */
} ** timed_task_handles = NULL;
int  timed_task_handle_count = 0;

static struct timed_task_handle ANY_TASK_dummy = {
    TIMED_TASK_HANDLE_magic,
    NULL,
    NO_timed_task_free,
    NO_inc_timed_task_refcount,
    NO_timed_task_action,

    {  /* NO_schedule_timelimit */
	0,
#ifdef USE_GETTIMEOFDAY
	{ 0 , 0 },
#endif
    
#ifdef USE_CLOCK_MONOTONIC
	{ 0 , 0 },
#endif
    },

    0,0,0
};

const struct timed_task_handle * ANY_TASK = &ANY_TASK_dummy;

static void free_timed_task_handle P_((struct timed_task_handle **handle,int found));
static void free_timed_task_handle(handle,found)
     struct timed_task_handle **handle;
     int found;
{
    if (TIMED_TASK_HANDLE_magic != (*handle)->magic)
	panic("SCHEDULE PANIC",__FILE__,__LINE__,
	      "free_timed_task_handle",
	      "Bad magic number",0);

    if (timed_task_handles) {

	if (found >= 0 && found < timed_task_handle_count &&
	    *handle == timed_task_handles[found]) {

	    DPRINT(Debug,9,(&Debug,
			    "free_timed_task_handle: timed task %d -- task handle %p\n",
			    found,*handle));
	    
	} else {
	    
	    int i;
	    found = -1;
	    
	    for (i = 0; i < timed_task_handle_count; i++) {
		if (*handle == timed_task_handles[i]) {
		    found = i;
		    
		    DPRINT(Debug,9,(&Debug,
				    "free_timed_task_handle: timed task %d -- task handle %p\n",
				    i,*handle));
		}
	    }
	}
    } else {
	found = -1;
    }
    
    if ((*handle)->busy) {
	
	if ((*handle)->delete)
	    panic("SCHEDULE PANIC",__FILE__,__LINE__,
		  "free_timed_task_handle",
		  "Busy and delete set",0);

	DPRINT(Debug,9,(&Debug,
			"free_timed_task_handle: busy, delete pending (task handle %p)\n",
			*handle));
	    
	(*handle)->delete = 1;
	(*handle) = NULL;
	return;
    }

    if ((*handle)->delete) {
	DPRINT(Debug,9,(&Debug,
			"free_timed_task_handle: delete pending -- task handle %p\n",
			*handle));
    } else if (-1 == found) {
	panic("SCHEDULE PANIC",__FILE__,__LINE__,
	      "free_timed_task_handle",
	      "Not on timed_task_handles",0);
    }

    if ((*handle)->task)
	(*handle)->free_task(&  ((*handle)->task));

    if ((*handle)->task) {
	DPRINT(Debug,9,(&Debug,
			"free_timed_task_handle: [%d], task %p not free'ed,  task handle %p\n",
			found,(*handle)->task,*handle));
    }

    (*handle)->deadline = NO_schedule_timelimit;

    (*handle)->magic = 0;  /* Invalidate */
    free(*handle);
    *handle = NULL;

    /* If handle != &timed_task_handles[found] 
       clear pointer now, when it is not cleared above
    */
    
    if (found >= 0 && timed_task_handles[found]) {
	DPRINT(Debug,9,(&Debug,
			"free_timed_task_handle: clearing timed task %d -- task handle %p\n",
			found,timed_task_handles[found]));
	
	timed_task_handles[found] = NULL;
    }
}

static void remove_timed_task P_((int ptr));
static void remove_timed_task(ptr)
     int ptr;
{
    int i;
    
    if (ptr < 0 || ptr >=  timed_task_handle_count)
	panic("SCHEDULE PANIC",__FILE__,__LINE__,
	      "remove_timed_task",
	      "Bad index",0);

    DPRINT(Debug,4,(&Debug,
		    "** Removing timed task %d  (task handle %p)\n",
		    ptr,timed_task_handles[ptr]));

    
    if (timed_task_handles[ptr])
	free_timed_task_handle(& (timed_task_handles[ptr]),ptr);

    for (i = ptr+1; i < timed_task_handle_count; i++)
	timed_task_handles[i-1] = timed_task_handles[i];

    timed_task_handle_count--;
}

static struct timed_task_handle * malloc_timed_task_handle P_((int *ptr));
static struct timed_task_handle * malloc_timed_task_handle(ptr)
     int *ptr;
{
    struct timed_task_handle * ret = safe_zero_alloc(sizeof (*ret));

    ret->task              = NULL;
    ret->free_task         = NO_timed_task_free;
    ret->inc_task_refcount = NO_inc_timed_task_refcount;
    ret->task_action       = NO_timed_task_action;
    ret->deadline          = NO_schedule_timelimit;

    ret->fired  = 0;
    ret->busy   = 0;
    ret->delete = 0;

    ret->magic = TIMED_TASK_HANDLE_magic;
    
    if (*ptr >= 0 && *ptr < timed_task_handle_count &&
	timed_task_handles && !timed_task_handles[*ptr]) {

	DPRINT(Debug,9,(&Debug,
			"malloc_timed_task_handle: reusing timed task %d -- task handle %p\n",
			*ptr,ret));
	
	timed_task_handles[*ptr] = ret;
    } else {

	timed_task_handles =
	    safe_array_realloc(timed_task_handles, timed_task_handle_count+1,
			       sizeof (timed_task_handles[0]));

	timed_task_handles[timed_task_handle_count] = ret;
	*ptr = timed_task_handle_count;
	timed_task_handle_count++;
	
	DPRINT(Debug,9,(&Debug,
			"malloc_timed_task_handle: Adding timed task %d -- task handle %p\n",
			*ptr,ret));

    }
    
    return ret;
}

/* Return 1 if timed_task is sceduled 
   Return 0 if there is no POLL_method
*/
int alter_timed_task(handle,task,free_task,inc_task_refcount,task_action,deadline)
     struct timed_task_handle       ** handle;
     struct timed_task               * task   /* may be NULL */;   
     timed_task_free_f               * free_task;
     inc_timed_task_refcount_f       * inc_task_refcount;
     timed_task_action_f             * task_action;
     const struct schedule_timelimit * deadline;
{
    int ret = 0;
    int found = -1;
    
    if (*handle) {
	if (TIMED_TASK_HANDLE_magic != (*handle)->magic)
	    panic("SCHEDULE PANIC",__FILE__,__LINE__,
		  "alter_timed_task",
		  "Bad magic number",0);

	if (timed_task_handles) {
	    int i;
	    
	    for (i = 0; i < timed_task_handle_count; i++) {
		if (*handle == timed_task_handles[i]) {
		    found = i;
		    
		    DPRINT(Debug,9,(&Debug,
				    "alter_timed_task: timed task %d -- task handle %p\n",
				    i,*handle));
		}
	    }
	}	   	
    }

    
    if (NULL                       == task &&
	NO_timed_task_free         == free_task &&
	NO_inc_timed_task_refcount == inc_task_refcount &&
	NO_timed_task_action       == task_action) {
	
	DPRINT(Debug,9,(&Debug,
			"alter_timed_task: Timed task removal request\n"));

    remove_task:
	/*  free_timed_task_handle() sets (*handle)->delete if (*handle)->busy is set */
	
	if (*handle)
	    free_timed_task_handle(handle,found);

	DPRINT(Debug,9,(&Debug,
			"alter_timed_task=%d (delete)\n",ret));
	
	return ret;
    }

    if (!schedule_have_timelimit(deadline)) {
	DPRINT(Debug,9,(&Debug,
			"alter_timed_task: No deadline\n"));

	goto remove_task;
    }
    
    if (*handle && -1 == found)
	panic("SCHEDULE PANIC",__FILE__,__LINE__,
	      "alter_timed_task",
	      "Not on timed_task_handles",0);
    else if (!*handle) {

	if (POLL_method) {	
	    *handle = malloc_timed_task_handle(&found);

	    DPRINT(Debug,9,(&Debug,
			    "alter_timed_task: timed task %d -- task handle %p\n",
			    found,*handle));
	} else {
	    DPRINT(Debug,9,(&Debug,
			    "alter_timed_task=%d (No POLL_method)\n",
			    ret));

	    return ret;
	}
    } 

    if ((*handle)->task != task) {
	if (*handle && (*handle)->task)
	    (*handle)->free_task(& ((*handle)->task));
	
	if (*handle && (*handle)->task) {
	    DPRINT(Debug,9,(&Debug,
			    "alter_timed_task: Prevous task not free'ed() !\n"));
	}

	(*handle)->task = task;
	if ((*handle)->task)
	    inc_task_refcount((*handle)->task);
    }

    (*handle)->free_task         = free_task;
    (*handle)->inc_task_refcount = inc_task_refcount;
    (*handle)->task_action       = task_action;    
    (*handle)->deadline          = *deadline;
    (*handle)->fired             = 0;
    ret = 1;

    DPRINT(Debug,9,(&Debug,
		    "alter_timed_task=%d\n",
		    ret));
    
    return ret;
}

void NO_timed_task_free(task)
     struct timed_task  ** task;
{
    DPRINT(Debug,4,(&Debug,"NO_timed_task_free:  timed_task %p\n",
		    *task));

    *task = NULL;
}

void NO_inc_timed_task_refcount(task)
     struct timed_task   * task;
{
    DPRINT(Debug,4,(&Debug,"NO_inc_timed_task_refcount:  timed_task %p\n",
		    task));
}


enum timed_task_r NO_timed_task_action(task,now)
     struct timed_task   * task;
     const struct schedule_timelimit * now;
{
    DPRINT(Debug,4,(&Debug,"NO_timed_task_action:  timed_task %p\n",
		    task));

    return timed_task_disable;
}


/* -------------------------------------------------------------------- */


int have_actions()
{
    int count = actions_count + timed_task_handle_count;

    return count;
}


static int real_wait P_((action_routine                  * action,
			 const struct schedule_timelimit * wait_timeout,
			 const struct schedule_timelimit * next_timeout,
			 struct cancel_data              * cd,
			 struct timed_task_handle        * task,
			 struct schedule_timelimit       * resulttime
			 ));

static void real_wait_check_timeouts P_((const struct schedule_timelimit * current,
					 int                             * done_p,
					 int                             * result_p,
					 action_routine                  * action,
					 struct timed_task_handle        * task));
static void real_wait_check_timeouts(current,done_p,result_p,action,task)
     const struct schedule_timelimit * current;
     int                             * done_p;
     int                             * result_p;
     action_routine                  * action;
     struct timed_task_handle        * task;
{
    int result = *result_p;
    int done   = *done_p;
    int i;
    int empty_slot = 0;

    
    DPRINT(Debug,20,(&Debug,
		     "real_wait_check_timeouts: %d actions\n",
		     actions_count));
    
    for (i = 0; i < actions_count; i++) {

	if (schedule_timeout_reached(& (actions[i].next_timeout),
				     current)) {
	    
	    if (actions[i].busy) {
		DPRINT(Debug,20,(&Debug,
				 "real_wait_check_timeouts: %d busy (action %d) -- timout skipped\n",
				 actions[i].fd,i));
	    } else {		
		enum action_status	 val;
		actions[i].busy = 1;
		
		DPRINT(Debug,20,(&Debug,
				 "real_wait_check_timeouts: running timeout_act for %d (action %d)\n",
				 actions[i].fd,i));
		val = actions[i].timeout_act(actions[i].fd,
					     actions[i].data,
					     current);
		DPRINT(Debug,20,(&Debug,
				 "real_wait_check_timeouts: [%d].timeout_act=%d",i,val));
		if (action == actions[i].timeout_act ||
		    action == ANY_ACTION) {
		    done = 1;
		    result = 1;
		    DPRINT(Debug,20,(&Debug," <wait done> "));
		}
		switch (val) {
		case action_disable:
		    actions[i].timeout_act  = no_action_routine;
		    actions[i].next_timeout = NO_schedule_timelimit;
		    DPRINT(Debug,20,(&Debug," (action_disable)"));
		    break;
		case action_continue:
		    DPRINT(Debug,20,(&Debug," (action_continue)"));
		    break;			    			    
		}
		DPRINT(Debug,20,(&Debug,"\n"));

		if (no_action_routine != actions[i].timeout_act) {
		    
		    if (! schedule_set_next_timeout0(& actions[i].next_timeout,
						     current,
						     actions[i].timeout_sec,

						     se_use_time_MAX
						     )) {
			DPRINT(Debug,20,(&Debug, 
					 "real_wait_check_timeouts: [%d] time overflow or bad time\n",i));
			
		    }
		}		
		actions[i].busy = 0;
	    }
	}
    }

    DPRINT(Debug,20,(&Debug,
		     "real_wait_check_timeouts: %d timed tasks\n",
		     timed_task_handle_count));

    for (i = 0; i < timed_task_handle_count; i++) {
	if (timed_task_handles[i]) {
	    if (TIMED_TASK_HANDLE_magic != timed_task_handles[i]->magic)
		panic("SCHEDULE PANIC",__FILE__,__LINE__,
		      "real_wait_check_timeout",
		      "Bad magic number (timed_task_handle)",0);

	    if (timed_task_handles[i]->fired) {
		    DPRINT(Debug,20,(&Debug,
				     "real_wait_check_timeouts: timed task %d busy (task handle %p) -- already fired/activated or disabled\n",
				     i,timed_task_handles[i]));

	    } else if (schedule_timeout_reached(&(timed_task_handles[i]->deadline),
					 current)) {

		if (timed_task_handles[i]->busy) {
		    DPRINT(Debug,20,(&Debug,
				     "real_wait_check_timeouts: timed task %d busy (task handle %p) -- deadline skipped\n",
				     i,timed_task_handles[i]));

		} else {
		    enum timed_task_r r;
		    
		    timed_task_handles[i]->busy  = 1;
		    timed_task_handles[i]->fired = 1;
		    DPRINT(Debug,20,(&Debug,
				     "real_wait_check_timeouts: running task_action for timed task %d (task handle %p)\n",
				     i,timed_task_handles[i]));
		    
		    r = timed_task_handles[i]->task_action(timed_task_handles[i]->task,
							   current);

		    if (task == timed_task_handles[i] ||
			task == ANY_TASK) {
			done = 1;
			result = 1;
		    }

		    switch (r) {
		    case timed_task_default: DPRINT(Debug,20,(&Debug,
							      "real_wait_check_timeouts: timed task %d (task handle %p) action done%s%s\n",
							      i,timed_task_handles[i],
							      timed_task_handles[i]->fired  ? "" : ", rescheduled",
							      timed_task_handles[i]->delete ? ", delete pending" : ""));
			break;
		    case timed_task_retry:
			DPRINT(Debug,20,(&Debug,
					 "real_wait_check_timeouts: timed task %d (task handle %p) action rescheduling%s\n",
					 i,timed_task_handles[i],
					 timed_task_handles[i]->delete ? ", delete pending" : ""));
			timed_task_handles[i]->fired = 0;
			break;
		    case timed_task_disable:
			DPRINT(Debug,20,(&Debug,
					 "real_wait_check_timeouts: timed task %d (task handle %p) action disabling (marking as fired/activated)%s\n",
					 i,timed_task_handles[i],
					 timed_task_handles[i]->delete ? ", delete pending" : ""));
			timed_task_handles[i]->fired = 1;
			break;
		    }
		    
		    timed_task_handles[i]->busy = 0;
		    if (timed_task_handles[i]->delete) {
			DPRINT(Debug,20,(&Debug,
				     "real_wait_check_timeouts: timed task %d delete pending (task handle %p) -- removing\n",
				     i,timed_task_handles[i]));

			free_timed_task_handle(& (timed_task_handles[i]),i);
		    }
		}

	    }
	} else {
	    empty_slot++;
	}	
    }
    
    DPRINT(Debug,20,(&Debug,
		     "real_wait_check_timeouts: %d empty slots on timed tasks\n",
		     empty_slot));

    
    *result_p = result;
    *done_p   = done;
}


    
static void real_wait_check_endtimeout P_((const struct schedule_timelimit  *
					   current,
					   int *done_p,
					   int *result_p,
					   const struct schedule_timelimit  *
					   wait_timeout,
					   const struct schedule_timelimit *
					   next_timeout));
static void real_wait_check_endtimeout(current,done_p,result_p,
				       wait_timeout,next_timeout)
    const struct schedule_timelimit * current;
    int *done_p;
    int *result_p;
    const struct schedule_timelimit * wait_timeout;
    const struct schedule_timelimit * next_timeout;
{
    int result = *result_p;
    int done   = *done_p;

    if (schedule_timeout_reached(next_timeout,current)) {
	done = 1;
	if (result == 0)
	    result = -1;

	DPRINT(Debug,20,(&Debug,
			 "real_wait_check_endtimeout: next timeout reached\n"));
    }

    if (schedule_timeout_reached( wait_timeout,current)) {
	done = 1;
	result = 1;
	
	DPRINT(Debug,20,(&Debug,
			 "real_wait_check_endtimeout: wait timeout reached\n"));
    }
  
    *result_p = result;
    *done_p   = done;
}



#ifdef  HAVE_SELECT
/* --------------------- select() ------------------------------------------ */

static int real_wait_select P_((action_routine                  * action,
				const struct schedule_timelimit * wait_timeout,
				const struct schedule_timelimit * next_timeout,
				struct cancel_data              * cd,
				struct timed_task_handle        * task,
				/* Updates *resulttime */
				struct schedule_timelimit       * resulttime));
static int real_wait_select(action,wait_timeout,next_timeout,cd,task,resulttime)
    action_routine                  * action;
    const struct schedule_timelimit * wait_timeout;
    const struct schedule_timelimit * next_timeout;
    struct cancel_data              * cd;
    struct timed_task_handle        * task;
    /* Updates *resulttime */
    struct schedule_timelimit       * resulttime;

{
    int result = 0;
    int done = 0;
    int err = 0;
    char * WA, * NA;
    
    char *Y UNUSED_VAROK = "action";
    char *Z UNUSED_VAROK = "timed task";
    
    if (action == no_action_routine)
	Y = "no action";
    else if (action == ANY_ACTION)
	Y = "any action";
    if (!task)
	Z = "no timed task";
    if (task == ANY_TASK)
	Z = "any timed task";
	        
    WA = schedule_timeout_string(wait_timeout);
    NA = schedule_timeout_string(next_timeout);
    
    DPRINT(Debug,20,(&Debug,
		     "real_wait_select: %s %p or %s %p, wait_timeout=%s, next_timeout=%s (actions_count=%d, %d timed tasks)\n", 	       
		     Y,action,
		     Z,task,
		     WA ? WA : "???",
		     NA ? NA : "???",
		     actions_count,
		     timed_task_handle_count));

    if (WA) {
	free(WA);
	WA = NULL;
    }

    if (NA) {
	free(NA);
	NA = NULL;
    }

    do {
	int i, n = 0;
	fd_set read_set;
	fd_set write_set;
	fd_set exp_set;
	struct timeval timeout, *timeout_val = NULL;
	int ret;
	struct schedule_timelimit  current = NO_schedule_timelimit;
	char * current_s = NULL;
	int cancel_delay = 0;

	FD_ZERO(&read_set);
	FD_ZERO(&write_set);
	FD_ZERO(&exp_set);

	DPRINT(Debug,20,(&Debug,
			 "real_wait_select: actions:"));
	for (i = 0; i < actions_count; i++) {
	    if (actions[i].fd >= FD_SETSIZE) {
		DPRINT(Debug,20,(&Debug,
			     " (fd=%d TOO BIG, FDSETSIZE=%d)",
				 actions[i].fd,FD_SETSIZE));
		continue;
	    }
	    
	    if (n <= actions[i].fd)
		n = actions[i].fd+1;

	    DPRINT(Debug,20,(&Debug,
			     " (fd=%d",actions[i].fd));
	    FD_SET(actions[i].fd,&exp_set);

	    if (actions[i].busy) {
		DPRINT(Debug,20,(&Debug,
				 " busy!"));
	    } else {
		if (actions[i].read_act != no_action_routine) {
		    FD_SET(actions[i].fd,&read_set);
		    DPRINT(Debug,20,(&Debug,
				     " READ"));
		}
	    
		if (actions[i].write_act != no_action_routine) {
		    FD_SET(actions[i].fd,&write_set);
		    DPRINT(Debug,20,(&Debug,
				     " WRITE"));
		}	
		if (actions[i].timeout_act != no_action_routine) {
		    DPRINT(Debug,20,(&Debug,
				     " TIMEOUT"));
		}
	    }
	    DPRINT(Debug,20,(&Debug, ")"));
	}
	DPRINT(Debug,20,(&Debug, "\n"));	
	
	if (schedule_timeout_to_timeval(next_timeout,&timeout)) {

	    DPRINT(Debug,20,(&Debug,
			     "real_wait_select: diff=%ld.%06ld\n",
			     (long)timeout.tv_sec,
			     timeout.tv_usec));	    
	    timeout_val        = &timeout;
	}

	if (cd) {
	    int delay_msec;

	    if (is_schedule_cancel(cd,&delay_msec)) {
		
		DPRINT(Debug,20,(&Debug, "real_wait_select: schedule cancel %d milliseconds\n",
				 delay_msec));

		if (delay_msec < 0)
		    delay_msec = 0;

		if (! timeout_val ||
		    timeout_val->tv_sec > delay_msec/1000
		    ||
		    ( timeout_val->tv_sec == delay_msec/1000 &&
		      timeout_val->tv_usec > delay_msec % 1000 )
		    ) {

		    timeout.tv_usec = 1000 * (  delay_msec % 1000 );
		    timeout.tv_sec  = delay_msec/1000;

		    cancel_delay = 1;
		    timeout_val        = &timeout;
		}
	    }	    
	}

	DPRINT(Debug,20,(&Debug, "real_wait_select: n = %d, read set=", n));
	for (i = 0; i < n; i++) {
	    if (FD_ISSET(i,&read_set)) {
		DPRINT(Debug,20,(&Debug, " %d", i));
	    }
	}
	DPRINT(Debug,20,(&Debug, ", write set="));
	for (i = 0; i < n; i++) {
	    if (FD_ISSET(i,&write_set)) {
		DPRINT(Debug,20,(&Debug, " %d", i));
	    }
	}
	DPRINT(Debug,20,(&Debug,", exception set="));
	for (i = 0; i < n; i++) {
	    if (FD_ISSET(i,&exp_set)) {
		DPRINT(Debug,20,(&Debug, " %d", i));
	    }
	}
	DPRINT(Debug,20,(&Debug, ", timeout="));
	if (timeout_val) {
	    DPRINT(Debug,20,(&Debug, " (sec=%ld usec=%ld)\n", 
			     (long) timeout_val->tv_sec,
			     (long) timeout_val->tv_usec));
	} else {
	    DPRINT(Debug,20,(&Debug, "NULL\n")); 
	}

	wait_can_signal = 1;
#ifdef _HPUX_SOURCE
	ret = select(n,(int *)&read_set,(int *)&write_set,
		     (int *)&exp_set,timeout_val);
#else
	ret = select(n,&read_set,&write_set,&exp_set,timeout_val);
#endif
	wait_can_signal = 0;

	if (ret < 0) {
	    err = errno;
	    DPRINT(Debug,20,(&Debug,
			     "real_wait_select select: %s (errno %d)\n",
			     strerror(err),err));
	    result = 0;

	    if (EINTR == err && cd && is_canceled(cd)) {
		DPRINT(Debug,20,(&Debug,
				 "real_wait_select: wait canceled\n"));		    	
	    }

	    break;
	}

	schedule_set_current_time(&current);
	current_s = schedule_timeout_string(&current);
	DPRINT(Debug,20,(&Debug,
			 "real_wait_select: select=%d, current=%s\n",
			 ret,
			 current_s ? current_s : "??"));
	if (current_s) {
	    free(current_s);
	    current_s = NULL;
	}
	*resulttime = current;
	
#ifdef USE_DLOPEN
	if (ret > 0) {
	    /* If we just timed out, it is quite predictable
	       so no seeding 
	    */
	    union xxx_rand {
		struct collection {
		    struct timeval XX;   /* someone may modify this */
		    struct schedule_timelimit X;
		    fd_set read_set;
		    fd_set write_set;
		    int r;
		} X;
		char bytes [sizeof (struct collection)];
	    } A;

	    if (timeout_val)
		A.X.XX = *timeout_val;
	    A.X.X         = current;
	    memcpy( &(A.X.read_set), &(read_set), sizeof read_set);
	    memcpy( &(A.X.write_set), &(write_set), sizeof write_set);
	    A.X.r         = ret;

	    seed_rand_bits(A.bytes, sizeof A, 
			   4 /* entropy bits */ );
	}
#endif

	if (0 == ret && cancel_delay && cd) 
	    schedule_cancel_timeout(cd);
	
	if (ret > 0) {
	    for (i = 0; i < actions_count; i++) {
		if (actions[i].fd >= FD_SETSIZE) {
		    DPRINT(Debug,20,(&Debug, 
				     "real_wait_select: fd=%d too big  FDSETSIZE=%d (action %d)\n",
				     actions[i].fd,FD_SETSIZE,i));
		    continue;
		}
		
		if (!actions[i].busy) {
		    actions[i].busy = 1;
		    if (FD_ISSET(actions[i].fd,&read_set) ||
			FD_ISSET(actions[i].fd,&exp_set)) {
			enum action_status val; 
			DPRINT(Debug,20,(&Debug, 
					 "real_wait_select: running read_act for %d (action %d)\n",
					 actions[i].fd,i));
			
			val  = actions[i].read_act(actions[i].fd,
						   actions[i].data,
						   &current);
			DPRINT(Debug,20,(&Debug,
					 "real_wait_select: [%d].read_act=%d",i,val));
			
			if (action == actions[i].read_act ||
			    action == ANY_ACTION) {
			    done = 1;
			    result = 1;
			    DPRINT(Debug,20,(&Debug," <wait done> "));
			}
			switch (val) {
			case action_disable:
			    actions[i].read_act = no_action_routine;
  			    DPRINT(Debug,20,(&Debug," (action_disable)"));
			    break;
			case action_continue:
			    DPRINT(Debug,20,(&Debug," (action_continue)"));
			    break;			    			    
			}
			DPRINT(Debug,20,(&Debug,"\n"));
		    }
		    
		    if (FD_ISSET(actions[i].fd,&write_set) ||
			FD_ISSET(actions[i].fd,&exp_set)) {
			enum action_status val;
			DPRINT(Debug,20,(&Debug,
					 "real_wait_select: running write_act for %d (action %d)\n",
					 actions[i].fd,i));
			val = actions[i].write_act(actions[i].fd,
						   actions[i].data,
						   &current);
			DPRINT(Debug,20,(&Debug,
					 "real_wait_select: [%d].write_act=%d",i,val));
			if (action == actions[i].write_act ||
			    action == ANY_ACTION) {
			    done = 1;
			    result = 1;
			    DPRINT(Debug,20,(&Debug," <wait done> "));
			}
			switch (val) {
			case action_disable:
			    actions[i].write_act = no_action_routine;
			    DPRINT(Debug,20,(&Debug," (action_disable)"));
			    break;
			case action_continue:
			    DPRINT(Debug,20,(&Debug," (action_continue)"));
			    break;		 
			}
			DPRINT(Debug,20,(&Debug,"\n"));
		    }
		    actions[i].busy = 0;
		}
		if (FD_ISSET(actions[i].fd,&exp_set)) {
		    DPRINT(Debug,20,(&Debug,
				     "real_wait_select: [%d] exception\n",i));
		    actions[i].write_act = no_action_routine;
		    actions[i].read_act = no_action_routine;
		    done = 1;
		    result = 1;
		}
	    }
	}

	real_wait_check_timeouts(&current,&done,&result,action,task);

	real_wait_check_endtimeout(&current,&done,&result,
				   wait_timeout,next_timeout);

	DPRINT(Debug,20,(&Debug,
			 "real_wait_select: done=%d, result=%d\n",done,result));

    } while(!done);

    
    DPRINT(Debug,20,(&Debug,
		     "real_wait_select=%d (errno=%d)\n",result,err));
    errno = err;
    return result;
}

#endif


#ifdef POLL
/* --------------------- poll() -------------------------------------------- */
 
static int real_wait_poll P_((action_routine                  * action,
			      const struct schedule_timelimit * wait_timeout, 
			      const struct schedule_timelimit * next_timeout,
			      struct cancel_data              * cd,
			      struct timed_task_handle        * task,
			      /* Updates *resulttime */
			      struct schedule_timelimit       * resulttime
			      ));

static int real_wait_poll(action,wait_timeout,next_timeout,cd,task,resulttime)
    action_routine                  * action;
    const struct schedule_timelimit * wait_timeout;
    const struct schedule_timelimit * next_timeout;
    struct cancel_data              * cd;
    struct timed_task_handle        * task;
    /* Updates *resulttime */
     struct schedule_timelimit       * resulttime;
{
    int result = 0;
    int done = 0;
    int err = 0;
    struct pollfd  * polls = NULL;
    int polls_count = 0;
    char * WA, * NA;
    char *Y UNUSED_VAROK = "action";
    char *Z UNUSED_VAROK = "timed task";
     
    if (action == no_action_routine)
	Y = "no action";
    else if (action == ANY_ACTION)
	Y = "any action";
    if (!task)
	Z = "no timed task";
    if (task == ANY_TASK)
	Z = "any timed task";
    
    WA = schedule_timeout_string(wait_timeout);
    NA = schedule_timeout_string(next_timeout);
    
    DPRINT(Debug,20,(&Debug,
		     "real_wait_poll: %s %p or %s %p, wait_timeout=%s, next_timeout=%s (actions_count=%d, %d timed tasks)\n",
		     Y,action,
		     Z,task,
		     WA ? WA : "???",
		     NA ? NA : "???",
		     actions_count,
		     timed_task_handle_count));

    if (WA) {
	free(WA);
	WA = NULL;
    }

    if (NA) {
	free(NA);
	NA = NULL;
    }


    do {
	int i;
	int ret;
	struct schedule_timelimit  current = NO_schedule_timelimit;
	char * current_s = NULL;
	int timeout = -1;
	int cancel_delay = 0;

	if (polls_count != actions_count) {
	    polls = safe_array_realloc(polls,
				       actions_count, sizeof(struct pollfd));
	    polls_count = actions_count;
	    bzero(polls,actions_count * sizeof(struct pollfd));
	    DPRINT(Debug,20,(&Debug,
			     "real_wait_poll: polls=%p, polls_count=%d\n",
			     polls,polls_count));	    
	}


	DPRINT(Debug,20,(&Debug,
			 "real_wait_poll: actions:"));
	for (i = 0; i < actions_count; i++) {

	    DPRINT(Debug,20,(&Debug, " (fd=%d",actions[i].fd));
	    polls[i].fd     = actions[i].fd;
	    polls[i].events = 0;
	    polls[i].revents = 0;

	    if (actions[i].busy) {
		DPRINT(Debug,20,(&Debug, " busy!"));
	    } else {
		if (actions[i].read_act != no_action_routine) {
		    polls[i].events |= POLLIN | POLLHUP;
		    DPRINT(Debug,20,(&Debug, " READ"));
		}
	    
		if (actions[i].write_act != no_action_routine) {
		    polls[i].events |= POLLOUT;
		    DPRINT(Debug,20,(&Debug, " WRITE"));
		}	
		if (actions[i].timeout_act != no_action_routine) {
		    DPRINT(Debug,20,(&Debug, " TIMEOUT"));
		}
	    }
	    DPRINT(Debug,20,(&Debug, ")"));
	}
        DPRINT(Debug,20,(&Debug, "\n"));	

	if (schedule_timeout_to_timeout_msec(next_timeout,
					     &timeout)) {
					    
	    DPRINT(Debug,20,(&Debug, "real_wait_poll: diff=%d.%03d\n",
			     timeout/1000,timeout%1000));
	    
	}

	if (cd) {
	    int delay_msec;

	    if (is_schedule_cancel(cd,&delay_msec)) {
		
		DPRINT(Debug,20,(&Debug, "real_wait_poll: schedule cancel %d milliseconds\n",
				 delay_msec));
		
		if (delay_msec < 0)
		    delay_msec = 0;

		if (timeout > delay_msec) {
		    cancel_delay = 1;
		    timeout = delay_msec;		    
		}
	    }
	}
	
	DPRINT(Debug,20,(&Debug,"real_wait_poll: polls_count = %d, polls=", 
			 polls_count));
	for (i = 0; i < polls_count; i++)
	    DPRINT(Debug,10,(&Debug, " {fd=%d events=%d}", 
			     polls[i].fd, polls[i].events));


	DPRINT(Debug,20,(&Debug, ", timeout=%d\n", 
			 timeout));

	wait_can_signal = 1;
	ret = poll(polls,polls_count,timeout);
	wait_can_signal = 0;

	if (ret < 0) {
	    err = errno;
	    DPRINT(Debug,20,(&Debug,
			"real_wait_poll poll: %s (errno %d)\n",
			strerror(err),err));
	    result = 0;

	    if (EINTR == err) {
		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,20,(&Debug,
				     "real_wait_poll: wait canceled\n"));		    
		}
	    }
	    
	    break;
	}

	schedule_set_current_time(&current);
	current_s = schedule_timeout_string(&current);
	DPRINT(Debug,20,(&Debug, "real_wait_poll: poll=%d, current=%s\n",
			 ret,
			 current_s ? current_s : "??"));
	if (current_s) {
	    free(current_s);
	    current_s = NULL;
	}
	*resulttime = current;

#ifdef USE_DLOPEN
	if (ret > 0) {
	    /* If we just timed out, it is quite predictable
	       so no seeding 
	    */
	    union xxx_rand {
                struct collection {
		    int    r;
		    struct schedule_timelimit t;
		} X;
		char bytes [sizeof (struct collection)];
	    } A;

	    A.X.r = ret;
	    A.X.t = current;

	    seed_rand_bits(A.bytes, sizeof A, 
			   4 /* entroby bits */ );

	}
#endif

	if (0 == ret && cancel_delay && cd) 
	    schedule_cancel_timeout(cd);

	if (ret > 0) {
	    for (i = 0; i < polls_count && i < actions_count; i++) {
		if (polls[i].revents) {
		    DPRINT(Debug,20,(&Debug, 
				     "real_wait_poll: polls[%d]: fd=%d revents=%d\n",
				     i,polls[i].fd,polls[i].revents));
		}
		if (!actions[i].busy) {
		    actions[i].busy = 1;
		    if (0 != (polls[i].revents & (POLLIN|POLLHUP|POLLERR))) {
			enum action_status val; 
			DPRINT(Debug,20,(&Debug,
					 "real_wait_poll: running read_act for %d (action %d)\n",
					 actions[i].fd,i));
			val  = actions[i].read_act(actions[i].fd,
						   actions[i].data,
						   &current);
		       DPRINT(Debug,20,(&Debug,
					"real_wait_poll: [%d].read_act=%d",i,val));
			if (action == actions[i].read_act ||
			    action == ANY_ACTION) {
			    done = 1;
			    result = 1;
			     DPRINT(Debug,20,(&Debug," <wait done> "));
			}
			switch (val) {
			case action_disable:
			    actions[i].read_act = no_action_routine;
			    DPRINT(Debug,20,(&Debug," (action_disable)"));
			    break;
			case action_continue:
			    DPRINT(Debug,20,(&Debug," (action_continue)"));
			    break;		       
			}
			DPRINT(Debug,20,(&Debug,"\n"));
		    }
		    
		    if (0 != (polls[i].revents & (POLLOUT|POLLERR))) {
			enum action_status val;
			DPRINT(Debug,20,(&Debug,
					 "real_wait_poll: running write_act for %d (action %d)\n",
					 actions[i].fd,i));
			val = actions[i].write_act(actions[i].fd,
						   actions[i].data,
						   &current);
			DPRINT(Debug,20,(&Debug,
					 "real_wait_poll: [%d].write_act=%d",i,val));
			if (action == actions[i].write_act ||
			    action == ANY_ACTION) {
			    done = 1;
			    result = 1;
			    DPRINT(Debug,20,(&Debug," <wait done> "));
			}
			switch (val) {
			case action_disable:
			    actions[i].write_act = no_action_routine;
			    DPRINT(Debug,20,(&Debug," (action_disable)"));
			    break;
			case action_continue:
			    DPRINT(Debug,20,(&Debug," (action_continue)"));
			    break;			    	      
			}
			DPRINT(Debug,20,(&Debug,"\n"));
		    }
		    actions[i].busy = 0;
		}
		if (polls[i].revents & POLLERR) {
		    DPRINT(Debug,20,(&Debug,
				     "real_wait_poll: [%d] POLLERR\n",i));
		    actions[i].write_act = no_action_routine;
		    actions[i].read_act = no_action_routine;
		    done = 1;
		    result = 1;
		}
	    }
	}

	real_wait_check_timeouts(&current,&done,&result,action,task);

	real_wait_check_endtimeout(&current,&done,&result,
				   wait_timeout,next_timeout);
	
	DPRINT(Debug,20,(&Debug,
			 "real_wait_poll: done=%d, result=%d\n",done,result));

    } while(!done);

    
    if (polls)
	free(polls);

    DPRINT(Debug,20,(&Debug,
		     "real_wait_poll=%d (errno=%d)\n",result,err));
    errno = err;
    return result;
}

#endif

static int real_wait(action,wait_timeout,next_timeout,cd,task,resulttime)
     action_routine                  * action;
     const struct schedule_timelimit * wait_timeout;
     const struct schedule_timelimit * next_timeout;
     struct cancel_data              * cd;
     struct timed_task_handle        * task;
     /* Updates *resulttime */
     struct schedule_timelimit       * resulttime;
{
    int ret = 0;

    switch (POLL_method) {
	unsigned int sleep_sec;
	
#ifdef HAVE_SELECT
    case POLL_method_select:
	ret = real_wait_select(action,wait_timeout,
			       next_timeout,cd,task,
			       resulttime);
	break;
#endif
#ifdef POLL
    case POLL_method_poll:
	ret = real_wait_poll(action,wait_timeout,
			     next_timeout,cd,task,
			     resulttime);
	break;
#endif
    default:

	if (schedule_timeout_to_timeout_sec(next_timeout,
					    & sleep_sec)) {
	    
	    DPRINT(Debug,20,(&Debug,
			     "real_wait: not available, diff=%u\n",
			     sleep_sec));

	    if ( sleep_sec > 0)
		sleep(sleep_sec);
	}

	schedule_set_current_time(resulttime);
		
	errno = ENOSYS;
	ret = -1;
	DPRINT(Debug,20,(&Debug,
			 "real_wait=%d: not available, POLL_method=%d\n",
			 ret,POLL_method));
    }

    return ret;
}

/* ----------- GENERIC ---------------------------------------------------- */

static int wait_for_something P_((action_routine            * action,
				  int                         seconds,
				  struct cancel_data        * cd,
				  struct timed_task_handle  * task,
				  const struct schedule_timelimit * deadline,
				  /* Updates *resulttime */
				  struct schedule_timelimit *resulttime));
static int wait_for_something(action,seconds,cd,task,deadline,resulttime)
     action_routine           * action;
     int                        seconds; 
     struct cancel_data       * cd;
     struct timed_task_handle * task;
     const struct schedule_timelimit * deadline;
     struct schedule_timelimit * resulttime;
{
    
    struct schedule_timelimit next_timeout = NO_schedule_timelimit;
    struct schedule_timelimit wait_timeout = NO_schedule_timelimit;
    struct schedule_timelimit current      = NO_schedule_timelimit;
    int result,err = 0;
    int done = 0;
    int i;

    schedule_set_current_time(&current);
        
    if (seconds >= 0 &&
	schedule_set_next_timeout0(& wait_timeout,
				   &current,
				   seconds,
				   se_use_time_MAX)) {
	char * X = schedule_timeout_string(& wait_timeout);

	if (X) {
	    DPRINT(Debug,10,(&Debug,"wait_for_something: wait_timeout=%s\n",
			     X));
	    
	    free(X);
	}
    }

    if (schedule_have_timelimit(deadline) &&
	schedule_shorten_next_timeout(& wait_timeout,deadline)) {
	char * X = schedule_timeout_string(& wait_timeout);

	if (X) {
	    DPRINT(Debug,10,(&Debug,"wait_for_something: uses deadline, wait_timeout=%s\n",
			     X));
	    
	    free(X);
	}
    }
    
    do {
	int have_busy = 0;
	
	next_timeout = wait_timeout;
	
	for (i = 0; i < actions_count; i++) {
	    if (actions[i].busy) {
		have_busy = 1;
	    } else {
		actions[i].busy = 1;

		switch(actions[i].schedule_act(actions[i].fd,
					       actions[i].schedule_data,
					       &current)) {
		case schedule_remove:
		    DPRINT(Debug,4,(&Debug,
			      "wait_for_something: Schedule action %d (fd=%d): Remove pending\n",
			      i,actions[i].fd)); 

		    clear_action(i);
		    break;
		case schedule_reconsider:
		    DPRINT(Debug,4,(&Debug,
				    "wait_for_something: Schedule action %d (fd=%d): Reconsider pending\n",
				    i,actions[i].fd));

		    schedule_set_current_time(& next_timeout);
		    current = next_timeout;

		    break;
		case schedule_have_action:
		    DPRINT(Debug,4,(&Debug,
				    "wait_for_something: Schedule action %d (fd=%d): Action filled\n",
				    i,actions[i].fd));
		    schedule_set_current_time(& next_timeout);
		    current = next_timeout;
		    
		    if (ANY_ACTION == action ||
			( no_action_routine != action &&
			  (action == actions[i].read_act ||
			   action == actions[i].write_act ||
			   action == actions[i].timeout_act)
			  )
			) {
			DPRINT(Debug,4,(&Debug,
					"wait_for_something: Schedule action %d (fd=%d): Ready\n",
					i,actions[i].fd));
			done = 1;
		    }
                    break;
		default:
		    break;
		}

		actions[i].busy = 0;
	    }
	}
	
	for (i = 0; i < actions_count; i++) {
	    if (!actions[i].busy &&
		no_action_routine != actions[i].timeout_act) {

		if (schedule_shorten_next_timeout(&next_timeout,
						  & actions[i].next_timeout)) {
		    char * X = schedule_timeout_string(& next_timeout);

		    if (X) {
			DPRINT(Debug,10,(&Debug,
					 "wait_for_something: Action %d (fd=%d): setting or shortened next_timeout to %s\n",
					 i,actions[i].fd,X));			
			free(X);
		    }		    
		}
	    }
	}

	for (i = 0; i < timed_task_handle_count; i++) {
	     if (timed_task_handles[i]) {
		if (TIMED_TASK_HANDLE_magic != timed_task_handles[i]->magic)
		    panic("SCHEDULE PANIC",__FILE__,__LINE__,
			  "wait_for_something",
			  "Bad magic number (timed_task_handle)",0);

		if (timed_task_handles[i]->busy)
		    have_busy = 1;
		else if (NO_timed_task_action != timed_task_handles[i]->task_action) {
		    if (timed_task_handles[i]->fired) {
			DPRINT(Debug,10,(&Debug,
					 "wait_for_something: timed task %d (task handle %p): already fired/activated or disabled\n",
					 i,timed_task_handles[i]));
			
		    } else if (schedule_shorten_next_timeout(&next_timeout,
						      &(timed_task_handles[i]->deadline))) {

			char * X = schedule_timeout_string(& next_timeout);

			if (X) {
			    DPRINT(Debug,10,(&Debug,
					     "wait_for_something: timed task %d (task handle %p): setting or shortened next_timeout to %s\n",
					     i,timed_task_handles[i],X));			
			    free(X);
			}
		    }
		}

	    }
	}
	
	if (schedule_invalid_next_timeout(& next_timeout)){
	    char * X = schedule_timeout_string(& next_timeout);
	    
	    DPRINT(Debug,1,(&Debug,
			    "wait_for_something: Next timeout %s -- setting to current time\n",
			    X ? X : "n/a"));

	    if (X) {
		free(X);
		X = NULL;
	    }
	    schedule_set_current_time(& next_timeout);
	    current = next_timeout;
	}

	
	result = real_wait(action,&wait_timeout,&next_timeout,cd,task,

			   /* Updates current to time where wait ended */
			   
			   &current);
	err = errno;

	if (have_busy) {
	    DPRINT(Debug,4,(&Debug,
			    "wait_for_something: Was busy actions or timed tasks, removals skipped\n"));
	} else {

	    if (actions_count > 0) {
		int w;

		for (w = actions_count; w > 0; w=i) {
		    i = w-1;
		    
		    if (no_action_routine == actions[i].read_act &&
			no_action_routine == actions[i].write_act &&
			no_action_routine == actions[i].timeout_act &&
			actions[i].schedule_act == no_schedule_routine) {
			if (actions[i].busy) {
			    DPRINT(Debug,4,(&Debug,
					    "wait_for_something: Action %d busy (fd=%d): remove skipped\n", 
					    i, actions[i].fd)); 
			} else {		       
			    remove_action(i);
			    break;
			}
		    }
		}
	    }

	    if (timed_task_handle_count > 0) {
		int w;
		
		for (w = timed_task_handle_count; w > 0 ; w=i) {
		    i = w-1;
		    
		    if (!timed_task_handles[i]) {
			/* remove empty slots */
			
			remove_timed_task(i);
		    }
		}
	    }
	}
	
	if (EINTR == err && cd && is_canceled(cd)) {
	    DPRINT(Debug,4,(&Debug,"** waiting canceled on interrupt\n"));
	    break;
	}

	
    } while (-1 == result && !done);

    if (resulttime) {
	char * current_s =  schedule_timeout_string(&current);

	if (current_s) {
	    DPRINT(Debug,10,(&Debug,"wait_for_something: wait end time is %s\n",
			     current_s));
	    free(current_s);
	}
	
	*resulttime = current;
    }
    
    errno = err;
    return result;
}

/* Got alarm --
   Perhaps jumped out from wait_* routines() 
   or arrived via setjump
*/
void wait_got_jump()
{
    int i;
    
    DPRINT(Debug,4,(&Debug,
		    "wait_got_jump:      INTERRUPTED\n"));
    wait_can_signal = 0;

    for (i = 0; i < actions_count; i++) {

	if (actions[i].busy) {
	    DPRINT(Debug,20,(&Debug,
			     "wait_got_jump: action for %d (action %d) was busy, resetting\n",
			     actions[i].fd,i));
	    actions[i].busy = 0;
	}
    }

    for (i = 0; i < timed_task_handle_count; i++) {
	if (timed_task_handles[i]) {
	    if (TIMED_TASK_HANDLE_magic != timed_task_handles[i]->magic)
		panic("SCHEDULE PANIC",__FILE__,__LINE__,
		      "wait_got_jump",
		      "Bad magic number (timed_task_handle)",0);
	    
	    if (timed_task_handles[i]->busy) {
		DPRINT(Debug,20,(&Debug,
				 "wait_got_jump: timed task %d  (task handle %p) was busy, resetting\n",
				 i,timed_task_handles[i]));

		timed_task_handles[i]->busy = 1;
	    }
	}
    }
}

int wait_for_action(action)
     action_routine * action;
{
    int ret;

    DPRINT(Debug,4,(&Debug,
		    "wait_for_action:    START | action=%p\n",action)); 
    ret=wait_for_something(action,-1,NULL,NULL,&NO_schedule_timelimit,
			   NULL);
    DPRINT(Debug,4,(&Debug,
		    "wait_for_action=%d END (errno=%d)\n",
		    ret,errno)); 

    return ret;
}

int wait_for_action_c(action,cd)
     action_routine * action;
     struct cancel_data *cd;
{
    int ret;

    DPRINT(Debug,4,(&Debug,
		    "wait_for_action_c:    START | action=%p\n",action)); 
    ret=wait_for_something(action,-1,cd,NULL,&NO_schedule_timelimit,NULL);
    DPRINT(Debug,4,(&Debug,
		    "wait_for_action_c=%d END (errno=%d)\n",
		    ret,errno)); 

    return ret;
}


int wait_for_timeout(seconds)
     int seconds;
{
    int ret;

    DPRINT(Debug,4,(&Debug,
		    "wait_for_timeout:    START | seconds=%d\n",seconds)); 
    ret = wait_for_something(no_action_routine,seconds,NULL,NULL,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,
		    "wait_for_timeout=%d END (errno=%d) \n",
		    ret,errno)); 
    return ret;
}

int wait_for_timeout_c(seconds,cd)
     int seconds;
     struct cancel_data *cd;
{
    int ret;

    DPRINT(Debug,4,(&Debug,
		    "wait_for_timeout_c:    START | seconds=%d\n",seconds)); 
    ret = wait_for_something(no_action_routine,seconds,cd,NULL,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,
		    "wait_for_timeout_c=%d END (errno=%d) \n",
		    ret,errno)); 
    return ret;
}


int wait_for_action_or_timeout(action,seconds)
     action_routine * action;
     int seconds;
{
    int ret;

    DPRINT(Debug,4,(&Debug,
		    "wait_for_action_or_timeout:    START | action=%p seconds=%d\n",
		    action,seconds)); 
    ret = wait_for_something(action,seconds,NULL,NULL,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,
		    "wait_for_action_or_timeout=%d END (errno=%d)\n",
		    ret,errno)); 
    return ret;
}

int wait_for_action_or_timeout_c(action,seconds,cd)
     action_routine * action;
     int seconds;
     struct cancel_data *cd;
{
    int ret;

    DPRINT(Debug,4,(&Debug,
		    "wait_for_action_or_timeout_c:    START | action=%p seconds=%d\n",
		    action,seconds)); 
    ret = wait_for_something(action,seconds,cd,NULL,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,
		    "wait_for_action_or_timeout_c=%d END (errno=%d)\n",
		    ret,errno)); 
    return ret;
}

int wait_for_any_action()
{
    int ret;

    DPRINT(Debug,4,(&Debug,"wait_for_any_action:    START\n")); 
    ret=wait_for_something(ANY_ACTION,-1,NULL,NULL,&NO_schedule_timelimit,
			   NULL);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action=%d END (errno=%d)\n",
		    ret,errno)); 

    return ret;
}

int wait_for_any_action_settime(resulttime)
     struct schedule_timelimit *resulttime;
{
    int ret;

    DPRINT(Debug,4,(&Debug,"wait_for_any_action_settime:    START\n")); 
    ret=wait_for_something(ANY_ACTION,-1,NULL,NULL,&NO_schedule_timelimit,
			   resulttime);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_settime=%d END (errno=%d)\n",
		    ret,errno)); 

    return ret;
}



int wait_for_any_action_c(cd)
     struct cancel_data *cd;
{
    int ret;

    DPRINT(Debug,4,(&Debug,"wait_for_any_action_c:    START\n")); 
    ret=wait_for_something(ANY_ACTION,-1,cd,NULL,&NO_schedule_timelimit,
			   NULL);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_c=%d END (errno=%d)\n",
		    ret,errno)); 

    return ret;
}

/* Updates *resulttime */
int wait_for_any_action_settime_c(resulttime,cd)
     struct schedule_timelimit *resulttime;
     struct cancel_data *cd;
{
    int ret;

    DPRINT(Debug,4,(&Debug,"wait_for_any_action_settime_c:    START\n")); 
    ret=wait_for_something(ANY_ACTION,-1,cd,NULL,&NO_schedule_timelimit,
			   resulttime);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_settime_c=%d END (errno=%d)\n",
		    ret,errno)); 

    return ret;
}

int wait_for_any_action_or_timeout(seconds)
     int seconds;
{
    int ret;
    
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_or_timeout:    START | seconds=%d\n",seconds)); 
    ret = wait_for_something(ANY_ACTION,seconds,NULL,NULL,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_or_timeout=%d END (errno=%d) \n",
		    ret,errno)); 
    return ret;
}

int wait_for_any_action_or_timeout_c(seconds,cd)
     int seconds;
     struct cancel_data *cd;
{
    int ret;
    
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_or_timeout_c:    START | seconds=%d\n",seconds)); 
    ret = wait_for_something(ANY_ACTION,seconds,cd,NULL,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_or_timeout_c=%d END (errno=%d) \n",
		    ret,errno)); 
    return ret;
}




/* Return 1 if task occured */
int wait_for_task(task)
     struct timed_task_handle  * task;
{
    int ret;

    DPRINT(Debug,4,(&Debug,"wait_for_task:    START | timed_task_handle=%p",task));

    if (task) {
	if (TIMED_TASK_HANDLE_magic != task->magic)
	    panic("SCHEDULE PANIC",__FILE__,__LINE__,
		  "wait_for_task",
		  "Bad magic number",0);
 
	DPRINT(Debug,4,(&Debug," timed_task %p",
			task->task));
    }
    DPRINT(Debug,4,(&Debug,"\n"));
    
    ret = wait_for_something(ANY_ACTION,-1,NULL,task,&NO_schedule_timelimit,
			     NULL);
    DPRINT(Debug,4,(&Debug,"wait_for_task=%d END (errno=%d) \n",
		    ret,errno)); 

    return ret;
}

int wait_for_action_or_deadline_c(action,deadline,cd)
     action_routine                  * action;
     const struct schedule_timelimit * deadline;
     struct cancel_data *cd;
{
    int ret;
    char *X =  schedule_timeout_string(deadline);
    
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_or_deadline_c:    START | deadline=%s\n",
		    X ? X : "n/a"));
    ret = wait_for_something(ANY_ACTION,-1,cd,NULL,deadline,NULL);
    DPRINT(Debug,4,(&Debug,"wait_for_any_action_or_deadline_c=%d END (errno=%d) \n",
		    ret,errno)); 

    if (X)
	free(X);
    
    return ret;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
